/*
Type checking:
	A PCTS shall verify that an object is of the type defined in
	the POSIX standard by testing all the following assertions
	as applicable to the object:

		 - The object shall have the same size as an object of the
		   specified type (the C sizeof operator shall be used to
		   test this assertion).  This assertion shall apply to
		   members of structures.  It shall not apply to structure
		   objects.

		 - The object shall be assigned to another object of the
		   defined type.

		 - The object shall be compared to another object of the
		   defined type.  This assertion shall not apply to
		   structure objects.

		 - The address of the object shall be assigned to another
		   object of the type pointer to the defined type.  This
		   assertion shall not apply to function return values.

	The above testable assertions shall not generate compile-
	time error or warning messages.

*/

/*
TYP_CHK_FUNCT(x, typ);
	Provide type checking on functions of other than an integral type.
		Where:
				x	Item to be type checked
				typ  Required return type 
						
		Example:
				TYP_CHK_FUNCT(funct(abc), char);
*/
#define TYP_CHK_FUNCT(x, typ) \
	continue_on(ERROR); \
	testing("Assign of \"" # x "\" to variable of type \"" # typ ".\""); \
	start \
		{typ y;  if(0) y=x; } \
	finis \
	testing("Compare \"" # x "\" to variable of type \"" # typ ".\""); \
	start \
		{typ y;  if(0) if(y == x); } \
	finis


/*
TYP_CHK_ENTITY(x, typ);
	Provide type checking on entities of other than an integral type.
		Where:
				x	Item to be type checked
				typ  Required return type 
						
		Example:
				TYP_CHK_ENTITY(ccc, char);
*/
#define TYP_CHK_ENTITY(x, typ) \
	continue_on(ERROR); \
	TYP_CHK_FUNCT(x, typ) \
	testing("Sizeof \"" # x "\" should == sizeof \"" # typ ".\""); \
	start \
		if(sizeof(x) != sizeof(typ)) \
			error("%s\nExpecting: %d\nActual: %d", \
			  	"Sizeof(" # typ ") does not equal sizeof(" # x ")",	\
			  	sizeof(x), \
			  	sizeof(typ) ); \
	finis


/*
TYP_CHK_PTR_VOID(x, typ);
	This is a subset of TYP_CHK_FUNCT_PTR() without the size checking
	test to allow for type checking of "pointers to void".
*/
#define TYP_CHK_PTR_VOID(x, typ) \
	continue_on(ERROR); \
	testing("Assign of \"" # x "\" to variable of type \"" # typ " *.\""); \
	start \
		{typ  *y;  if(0) y = x; } \
	finis \
	testing("Compare \"" # x "\" to variable of type \"" # typ " *.\""); \
	start \
		{typ  *y;  if(0) if (y == x); } \
	finis


#define TYP_CHK_ENTITY_PTR TYP_CHK_FUNCT_PTR
/*
TYP_CHK_FUNCT_PTR;
	Provide assertion checks on functions that return a pointer.
		Where:
				x:   Item to be checked
				typ: Required return type (without `*') 

		Example:
				TYP_CHK_FUNCT_PTR(strcat(s1,s2), char); 
				WARNING: second argument can be only one token;
					DO NOT use something like
					TYP_CHK_FUNCT_PTR(strcat(s1,s2), struct tms); 
*/
#define TYP_CHK_FUNCT_PTR(x, typ) \
	continue_on(ERROR); \
	TYP_CHK_PTR_VOID(x, typ) \
	testing("Sizeof \"" # x "\" should == sizeof \"" # typ " *.\""); \
	start \
		if(sizeof(x) != sizeof(typ *)) \
			error("%s\nExpecting: %d\nActual: %d", \
			  	"Sizeof(" # typ " *) does not equal sizeof(" # x ")", \
			  	sizeof(x), \
			  	sizeof(typ *) ); \
	finis


/*
TYP_CHK_FUNCT_INT;
	Provide assertion checks on functions that return integral type.
		Where:
				x:   Function whose return type is to be checked
				typ: Required return type 
		Example:
				TYP_CHK_FUNCT_INT(read(fd, buf, cnt), int); 
*/
#define TYP_CHK_FUNCT_INT(x, typ) \
	continue_on(ERROR); \
	testing("Assign of \"" # x "\" to variable of type \"" # typ ".\""); \
	start \
		{typ y;  if(0) y=x; } \
	finis \
	testing("Compare \"" # x "\" to variable of type \"" # typ ".\""); \
	start \
		{typ y=1;  if(0) if(y==x); }\
	finis \
	testing("Negation of integral type \"" # x "\" is acceptable."); \
	start \
		{typ y; if(0) y = ~(x); } \
	finis
	

/*
TYP_CHK_ENTITY_INT;
	Provide type checking on entities that provide integrals.
		Where:
				x	Item to be type checked
				typ  Required return type
*/
#define TYP_CHK_ENTITY_INT(x, typ) \
	continue_on(ERROR); \
	TYP_CHK_FUNCT_INT(x, typ) \
	testing("Sizeof \"" # x "\" should == sizeof \"" # typ ".\"");  \
	start \
		if(sizeof(x) != sizeof(typ)) \
			error("%s\nExpecting: %d\nActual: %d", \
			  	"Sizeof(" # typ ") does not equal sizeof(" # x ")",	\
			  	sizeof(x), \
			  	sizeof(typ) ); \
	finis \
	testing("Assign of &\"" # x "\" to variable of type \"" # typ ".\"\n"); \
	start \
		{ typ *y;  y = &x; } \
	finis


/*
TYP_CHK_ENTITY_ARRAY;
	Provide assertion checks on entities of types which are arrays.
		Where:
				x:   Entity being checked
				typ: Type expected 
*/
#define TYP_CHK_ENTITY_ARRAY(x, typ, size) \
	continue_on(ERROR); \
	testing("Assign of \"" # x "\" to variable of type \"" # typ " " # x "[" # size "].\""); \
	start \
		{ \
			int z; \
			{typ y[size];  for(z=0;z<size;++z) y[z] = x[z]; } \
		} \
	finis \
	testing("Compare \"" # x "\" to variable of type \"" # typ " " # x "[].\""); \
	start \
		{ \
			int z; \
			{typ y[size];  for(z=0;z<size;++z) if (y[z] == x[z]); }\
		} \
	finis \
	testing("Sizeof \"" # x "\" should == sizeof \"" # typ " " # x "[" # size "].\"");  \
	start \
		if(sizeof(x) != size * sizeof(typ)) \
			error("%s\nExpecting: %d\nActual: %d", \
		  		"(Sizeof(" # typ ") * " # size ") does not equal sizeof(" # x ")", \
		  		sizeof(x), \
		  		(size * sizeof(typ)) ); \
	finis
/*
TYP_CHK_UNSIGNED_INTEGRAL(x);
	Provide type checking on unsigned integral type.
		Where:
				x	Item to be type checked
*/
#define	TYP_CHK_UNSIGNED_INTEGRAL(x) \
	continue_on(ERROR); \
	testing("Negation of value with integral type \"" # x "\" is acceptable.");\
	start \
		{ x z=1; unsigned long y; if (0) y = ~(z); } \
	finis \
	testing("Type of \"" # x "\" is not a type real."); \
	start \
		{x z=1; \
			if((z=1.5) != 1) \
			error("Type expected unsigned integral, type provided is real."); \
		} \
	finis \
	testing("Type of \"" # x "\" is unsigned."); \
	start \
		{	x y; \
			int	i; \
			i = sizeof(x); \
			switch(i) { \
				case 1:		/* Type is a char */ \
					if((y=255) != 255) \
						error("Type (\"" # x "\") is a signed char."); \
					break; \
				case 2:		/* Type is two bytes long */ \
					if((y=65535) != 65535) \
						error("Type \"" # x "\" is a signed type."); \
					break; \
				case 4:		/* Type is four bytes long */ \
					if((y=4294967295) != 4294967295) \
						error("Type \"" # x "\" is a signed type."); \
					break; \
				default:		/* Type is larger than four bytes. */ \
					warning("Sign untested, type \"" # x "\" > four bytes."); \
					break; \
			} \
		} \
	finis
