#include <z_pcts.h>

#include <unistd.h>
#include <signal.h>
#include <element.h>

/*
 *	sv_setup.c
 *
 *	Called once for every executable which:
 *
 *		a.	uses machine dep constants
 *		b.	cares about signal catching
 *		c.	wants to use tmp/data directories
 *
 *	Prints out the name of the man section.
 *	Generates a warning for subsequent calls.
 *	Sets up a whole lot of signal handlers.
 *
 *	Note:
 *		Any changes to setup may also be needed in OS/exec
 *		which has fake setup so as not to destroy signal settings,
 *		close open file descriptors, etc, while testing exec.
 *
 */

struct siggs {
	int sigg;
	char *siggname;
	int sflg;
} siggs[] = {
		/* POSIX Required Signals */
#ifdef	SIGABRT
	{	SIGABRT,	"SIGABRT",	0	},
#endif
#ifdef	SIGALRM
	{	SIGALRM,	"SIGALRM",	0	},
#endif
#ifdef	SIGFPE
	{	SIGFPE,		"SIGFPE",	0	},
#endif
#ifdef	SIGHUP
	{	SIGHUP,		"SIGHUP",	0	},
#endif
#ifdef	SIGILL
	/*	Don't catch this non_POSIX signal, abort() currently needs this
		to function correctly in some implementations.
			{	SIGILL,		"SIGILL",	0	},
	*/
#endif
#ifdef	SIGINT
	{	SIGINT,		"SIGINT",	0	},
#endif
#ifdef	SIGPIPE
	{	SIGPIPE,	"SIGPIPE",	0	},
#endif
#ifdef	SIGQUIT
	{	SIGQUIT,	"SIGQUIT",	0	},
#endif
#ifdef	SIGSEGV
	{	SIGSEGV,	"SIGSEGV",	0	},
#endif
#ifdef	SIGTERM
	{	SIGTERM,	"SIGTERM",	0	},
#endif
#ifdef	SIGUSR1
	{	SIGUSR1,	"SIGUSR1",	0	},
#endif
#ifdef	SIGUSR2
	{	SIGUSR2,	"SIGUSR2",	0	},
#endif

/* SIGCONT catching function behavior is implementation defined	*/
/* Do not catch SIGCONT	*/

#ifdef	SIGTSTP
	{	SIGTSTP,	"SIGTSTP",	0	},
#endif
#ifdef	SIGTTIN
	{	SIGTTIN,	"SIGTTIN",	0	},
#endif
#ifdef	SIGTTOU
	{	SIGTTOU,	"SIGTTOU",	0	},
#endif

	/* Attempt to catch all signals which can occur on a POSIX
	   implementation which may provide additional signals.
	*/
/*
#ifdef	SIGTRAP
	{	SIGTRAP,	"SIGTRAP",	0	},
#endif
*/
#ifdef	SIGSYS
	{	SIGSYS,		"SIGSYS",	0	},
#endif
#ifdef	SIGIOT
	{	SIGIOT,		"SIGIOT",	0	},
#endif
#ifdef	SIGEMT
	{	SIGEMT,		"SIGEMT",	0	},
#endif
#ifdef	SIGBUS
	{	SIGBUS,		"SIGBUS",	0	},
#endif
#ifdef	SIGPWR
	{	SIGPWR,		"SIGPWR",	0	},
#endif
};

# define NSIGGS	(sizeof(siggs)/sizeof(struct siggs))

static	char	*sv_nclose = NULL;
static	char	*sv_nsig = NULL;
static	struct	sigaction sact_env;

extern	void	sv_config(void);
extern	void	sv_nproc(void);
extern	int	sv_get_outfd(void);
char	*sig_lookup(int);
void	sv_catchsig(void (*)(int));


/*
** sig_handler
**	Handle unexpected signals.
**
** Arguments
**	sig	signal number sent by the system.
**
** Side Effects
**	Prints an error message.
**	Attempts to send a SIGKILL to any children that might be around.
**	Calls done.
*/

void sig_handler(int sig)
{
	 sigaction(sig, &sact_env, (struct sigaction  *) 0);
	 sv_prerr(sv_xenv->sv_xsev,
		"Unexpected signal occurred: %s",sig_lookup(sig));
}

void sv_no_sig(int sig)
{
	int i;

	for(i=0;i<NSIGGS;i++)
		if(sig == siggs[i].sigg)
			siggs[i].sflg=1;
}


void sv_setup(int e_index)
{
	int		i;
	char	*datadir, *tempdir;
	long	open_max;

	datadir = sv_data_path();
	tempdir = sv_temp_path();
	(void) strcpy(datadir,"../data/");
	(void) strcpy(tempdir,"../tmp/"); 
	(void) strcat(datadir, element[e_index].e_name);
	(void) strcat(tempdir, element[e_index].e_name);
	(void) strcat(datadir, ".d/");
 	(void) strcat(tempdir, ".d/");

	/*
 	* Set up installation dependent values for use in util.c
 	* functions.
 	*/
	sv_config();
	sv_nproc();
	sv_catchsig(sig_handler);

	/*
 	* close everything except stdin, stdout, stderr
 	*/
	errno = 0;					/* If value not defined, errno unchanged. */
	if((open_max = sysconf(_SC_OPEN_MAX)) == -1) {
		if(errno == 0)
			open_max = INT_MAX;		/* Allow for no limit */
		else
			severr("sysconf(_SC_OPEN_MAX) failed.");
	}
	for(i = 3; i < open_max; i++) {
		if((i != sv_get_outfd()) && (!sv_nclose || !(*(sv_nclose+i))))
			(void) close(i);
	}
	if(sv_nclose)
		free(sv_nclose);
}


void sv_catchsig(void (*handler)(int))
{
	struct siggs *p = &siggs[0];
	int i;

	sact_env.sa_handler = handler;
	sigemptyset(&sact_env.sa_mask);
	sact_env.sa_flags = 0;

	/*
	** Set the signal catching for all known signals.
	*/

	for(i=0;i<NSIGGS;i++,p++) {
		if(!siggs[i].sflg) {
			if(sigaction(p->sigg, &sact_env, (struct sigaction  *) 0) == -1) {
				if(sysconf(_SC_JOB_CONTROL) == (long) -1) {
					if((p->sigg == SIGTSTP) || (p->sigg == SIGTTIN) ||
												(p->sigg == SIGTTOU))
						continue;
				}
				error("Failed to set signal handler for %s", p->siggname);
			}
		}
	}
	if(sv_nsig)
		free(sv_nsig);
}

/*
 *	sig_lookup
 *		Look up the symbolic name for a signal.
 */
char *sig_lookup(int signum)
{
	register int i;
	register struct siggs *s = &siggs[0];
	static char buf[32];

	for(i=0; i<NSIGGS; i++, s++)
		if(s->sigg == signum)
			return(s->siggname);
	(void) sprintf(buf,"%d (unknown signal)",signum);
	return(buf);
}
