/******************************************************************************
* termios.c
*	termios-related routines
******************************************************************************/
#include <pcts.h>

#include <Zunistd.h>
#include <sys/Zstat.h>
#include <Zfcntl.h>
#include <Ztermios.h>

#include <sv_env.h>
#include <sv_errno.h>

#include <pcts_timer.h>
#include <tssf.h>

const cc_t PERASE	= 0177;
const cc_t PKILL	= 0025;
const cc_t PINTR	= 0003;
const cc_t PQUIT	= 0034;
const cc_t PSTART	= 0021;
const cc_t PSTOP	= 0023;
const cc_t PEOF		= 0004;
const cc_t PEOL		= 0013;
const cc_t PSUSP	= 0032;
			/* Non-canonical read what's in input queue or that requested. */
const cc_t PMIN		= 0;
const cc_t PTIME	= 0;

/*********************************
* Baud rate table containing
* mask and associated decimal
* value
*********************************/
struct mask_table
{
	const speed_t 	sym_val;
	const char		*symbol;
	const int		value;
} masks[] =
{
	{ B50,	"B50",		50 },
	{ B75,	"B75",		75 },
	{ B110,	"B110",		110 },
	{ B134,	"B134",		134 },
	{ B150,	"B150",		150 },
	{ B200,	"B200",		200 },
	{ B300,	"B300",		300 },
	{ B600,	"B600",		600 },
	{ B1200,	"B1200",	1200 },
	{ B1800,	"B1800",	1800 },
	{ B2400,	"B2400",	2400 },
	{ B4800,	"B4800",	4800 },
	{ B9600,	"B9600",	9600 },
	{ B19200,	"B19200",	19200 },
	{ B38400,	"B38400",	38400 }
};
#define BAUD_CNT	(sizeof(masks)/sizeof(struct mask_table))

			/*	Return baud rate value given a symbolic baud rate. */
int baud_value(speed_t speed)
{ 
	int i;

	for(i=0; i<BAUD_CNT; ++i) {
		if(speed == masks[i].sym_val)
			return(masks[i].value);
	}
	error("Illegal value for GTI_SPEED: %s", speed);		/* no match */
}

			/*	Return baud rate mask for symbol string. */
speed_t baud_mask(const char *speed)
{ 
	int i;

	for(i=0; i<BAUD_CNT; ++i) {
		if(strcmp(speed, masks[i].symbol) == 0) {
			return(masks[i].sym_val);
		}
	}
	/* no match */
	error("Illegal value for GTI_SPEED: %s", speed);
}


		/*	Return the symbolic value for maximum baud rate
			specified in ./include/config
		*/ 
speed_t baud_max(void)
{
	int		sp_cur, sp_max;
	char	*p_cur, *p_max;
	char	rates[200];

		strcpy(rates, GTI_BAUD_RATES);
		if((p_max = strtok(rates, ",")) == NULL)
			error("GTI_BAUD_RATES in config configured as \"\".");
		sp_max = baud_value(baud_mask(p_max));
		while((p_cur = strtok(NULL, ",")) != NULL) {
			if((sp_cur = baud_value(baud_mask(p_cur))) > sp_max) {
				sp_max = sp_cur;
				p_max = p_cur;
			}
		}
	return(baud_mask(p_max));
}


		/*	Return the symbolic value for minimum baud rate
			specified in ./include/config
		*/ 
speed_t baud_min(void)
{
	int		sp_cur, sp_min;
	char	*p_cur, *p_min;
	char	rates[200];

		strcpy(rates, GTI_BAUD_RATES);
		if((p_min = strtok(rates, ",")) == NULL)
			error("GTI_BAUD_RATES in config configured as \"\".");
		sp_min = baud_value(baud_mask(p_min));
		while((p_cur = strtok(NULL, ",")) != NULL) {
			if((sp_cur = baud_value(baud_mask(p_cur))) < sp_min) {
				sp_min = sp_cur;
				p_min = p_cur;
			}
		}
	return(baud_mask(p_min));
}

	/*	If the speed argument is specified in parameter GTI_BAUD_RATES in
		./install/config, then return TRUE otherwise return FALSE.
	*/
int baud_rate(speed_t speed)
{
	char	*p_str;
	char	rates[200];

		strcpy(rates, GTI_BAUD_RATES);
		if((p_str = strtok(rates, ",")) == NULL)
			error("GTI_BAUD_RATES in config configured as \"\".");
		do {
			if(speed == baud_mask(p_str))
				return(TRUE);
		} while((p_str = strtok(NULL, ",")) != NULL);
	return(FALSE);
}


			/*	Sleep for at least an intergral number of second to allow
				n characters to read or written at specified baud rate.
			*/
void baud_sleep(int n, speed_t baud)
{
	int	t;

	/*
	 * assume 10 bits per character.
	 * and it takes less than a second for a read() and write()
	 * system call to return.
	 */
	t = (n+5)/(baud * 10) + 1;
	SLEEP(t);
}


void termios_prep(int *tty_fildes_p, struct termios *tty_termios_p,
				int *loop_fildes_p, struct termios *loop_termios_p)
{
	Zsetsid();
				/* Establish PORT_TTY as the controlling terminal. */
	tty_prep(tty_fildes_p, tty_termios_p);
	loop_prep(loop_fildes_p, loop_termios_p);
			/* Allow time for ports to be established before continuing. */
	sleep(1);
}


void termios_nct(int *tty_fildes_p, struct termios *tty_termios_p,
				int *loop_fildes_p, struct termios *loop_termios_p)
{
		/* Establish PORT_TTY and PORT_LOOP as non-controlling terminals. */
	tty_prep(tty_fildes_p, tty_termios_p);
	loop_prep(loop_fildes_p, loop_termios_p);
			/* Allow time for ports to be established before continuing. */
	sleep(1);
}


void tty_prep(int *tty_fildes_p, struct termios *tty_termios_p)
{
	int flags;

	SET_TIMEOUT
					/*	Establish controlling terminal using
						implementation supplied procedure.
					*/
		*tty_fildes_p = open_ctty(PORT_TTY, O_RDWR | O_NONBLOCK, PERM_CTTY);
		if(*tty_fildes_p < 0)
			error("Couldn't open controlling terminal.");
		tty_init(*tty_fildes_p, tty_termios_p);
		flags = Zfcntl(*tty_fildes_p, F_GETFL);		/* Clear O_NONBLOCK flag. */
		Zfcntl(*tty_fildes_p, F_SETFL, flags & ~O_NONBLOCK);
	CLEAR_ALARM
}


void loop_prep(int *loop_fildes_p, struct termios *loop_termios_p)
{
	int flags;

	SET_TIMEOUT
		*loop_fildes_p = Zopen(PORT_LOOP, O_RDWR | O_NONBLOCK);
		loop_init(*loop_fildes_p, loop_termios_p);
		flags = Zfcntl(*loop_fildes_p, F_GETFL); /* Clear O_NONBLOCK flag. */
		Zfcntl(*loop_fildes_p, F_SETFL, flags & ~O_NONBLOCK);
	CLEAR_ALARM
}

		/*	Save the initial termios state
			Modify struct termios to NIST-PCTS values
			Modify struct termios for test values
		*/
void tty_init(int fildes, struct termios *termios_p)
{
	Ztcgetattr(fildes, termios_p);
			/*	Set default NIST-PCTS values  */
	termios_p->c_iflag &= ~(PARMRK|INLCR|INPCK|IGNCR|IXOFF|BRKINT|IXON);
	termios_p->c_iflag |= (IGNPAR|IGNBRK);
	termios_p->c_iflag &= ~(ISTRIP|ICRNL);
	termios_p->c_oflag &= ~OPOST;

					/*	When CLOCAL is supported, ensure line is dropped
						when modem disconnects.
					*/
	termios_p->c_cflag |= HUPCL;
					/*	Clear CLOCAL only if GTI_MODEM_CONTROL is supported. */
	if(GTI_MODEM_CONTROL == 0)
		termios_p->c_cflag |= CLOCAL;
	else
		termios_p->c_cflag &= ~CLOCAL;

	termios_p->c_lflag &= ~(ECHO|TOSTOP|IEXTEN|ECHONL|ECHOE);
										/* 	TTY_LOOP canonical.	*/
	termios_p->c_lflag |= (ISIG|ICANON|NOFLSH);

	termios_p->c_cflag |= CREAD;
	termios_p->c_cflag &= ~CSIZE;
	termios_p->c_cflag |= CS8;
	termios_p->c_cflag |= (PARODD|PARENB);		/* Generated odd parity. */

	termios_p->c_cc[VMIN] = PMIN;
	termios_p->c_cc[VTIME] = PTIME;
	termios_p->c_cc[VINTR] = PINTR;
	termios_p->c_cc[VQUIT] = PQUIT;
	termios_p->c_cc[VERASE] = PERASE;
	termios_p->c_cc[VKILL] = PKILL;
	termios_p->c_cc[VEOF] = PEOF;
	termios_p->c_cc[VEOL] = PEOL;
	termios_p->c_cc[VSUSP] = PSUSP;

		/*	START and STOP will not be set for those implementations
			that do not support START and STOP changing.
			To ensure these parameters are not inadvertantly changed by
			the PCTS these values are intialized for those implementations
			that allow START and STOP to be changed.
		 */
	termios_p->c_cc[VSTART] = PSTART;
	termios_p->c_cc[VSTOP] = PSTOP;

	Zcfsetispeed(termios_p, baud_max());
	Zcfsetospeed(termios_p, baud_max());
	Ztcsetattr(fildes, TCSAFLUSH, termios_p);

}


void loop_init(int fildes, struct termios *termios_p)
{
	cc_t		posix_vdisable;

	posix_vdisable = (cc_t) Zfpathconf(fildes, _PC_VDISABLE);
	Ztcgetattr(fildes, termios_p);

							/* 	PORT_LOOP noncanonical.	*/
	termios_p->c_iflag &= ~(PARMRK|ISTRIP|ICRNL|INLCR|INPCK|IGNCR|IXOFF|BRKINT|IXON);
	termios_p->c_iflag |=(IGNPAR|IGNBRK);
	termios_p->c_oflag &= ~OPOST;

					/*	When CLOCAL is supported, ensure line is dropped
						when modem disconnects.
					*/
	termios_p->c_cflag |= HUPCL;
					/*	Clear CLOCAL only if GTI_MODEM_CONTROL is supported. */
	if(GTI_MODEM_CONTROL == 0)
		termios_p->c_cflag |= CLOCAL;
	else
		termios_p->c_cflag &= ~CLOCAL;

	termios_p->c_cflag |= CREAD;
	termios_p->c_cflag &= ~CSIZE;
	termios_p->c_cflag |= CS8;
	termios_p->c_cflag |= (PARODD|PARENB);		/* Generated odd parity. */

	termios_p->c_lflag &= ~(ECHO|ISIG|ICANON|TOSTOP|IEXTEN|ECHONL|ECHOE);
	termios_p->c_lflag |=(NOFLSH);

	termios_p->c_cc[VINTR] = posix_vdisable;
	termios_p->c_cc[VQUIT] = posix_vdisable;
	termios_p->c_cc[VERASE] = posix_vdisable;
	termios_p->c_cc[VKILL] = posix_vdisable;
	termios_p->c_cc[VEOF] = posix_vdisable;
	termios_p->c_cc[VEOL] = posix_vdisable;
	termios_p->c_cc[VSUSP] = posix_vdisable;
	termios_p->c_cc[VMIN] = PMIN;
	termios_p->c_cc[VTIME] = PTIME;

		/*	START and STOP will not be set for those implementations
			that do not support START and STOP changing.
			To ensure these parameters are not inadvertantly changed by
			the PCTS these values are intialized for those implementations
			that allow START and STOP to be changed.
		 */
	termios_p->c_cc[VSTART] = PSTART;
	termios_p->c_cc[VSTOP] = PSTOP;

	Zcfsetispeed(termios_p, baud_max());
	Zcfsetospeed(termios_p, baud_max());

	Ztcsetattr(fildes, TCSAFLUSH, termios_p);
}
